import numpy as np
import os
import matplotlib.pyplot as plt
from KDEpy import FFTKDE
from skimage import filters

'''Takes two burst values and calculates whether they overlap or not'''
def overlap(burst1, burst2):
    if burst1[1]<burst2[0] or burst2[1]<burst1[0]:
        return False
    else:
        return True

def network_burst_detection(outputpath, wells, electrode_amnt, measurements, hertz, min_channels, threshold_method, plot_electrodes, savedata=True):
    spikepath=f'{outputpath}/spike_values'
    burstpath=f'{outputpath}/burst_values'

    min_channels=round(min_channels*electrode_amnt)
    #print(f"Minimal amount of channels required for network burst: {min_channels}")
    total_network_burst=[]
    participating_bursts=[]
    
    for well in wells:
        fig, ax = plt.subplots(4,1, sharex=True, gridspec_kw={'height_ratios': [3, 1, 1, 1]}, figsize=(16,9))
        ax[1].set_ylabel("Active\nbursts")
        ax[0].set_xlim([0,measurements/hertz])
        ax[0].set_ylim([0,electrode_amnt+1])
        ax[0].set_yticks(np.arange(1, electrode_amnt+1, 1))
        ax[3].set_xlabel("Time in seconds")
        ax[1].set_ylim([0, electrode_amnt+1])
        ax[0].set_ylabel("Electrode")
        well_spikes=[]
        burst_spikes=[]
        spikes_for_kde=np.array([])
        burst_spikes_for_kde=np.array([])
        for electrode in range(1,electrode_amnt+1):
            # Load in spikedata
            spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode}_spikes.npy')
            burstdata=np.load(f'{burstpath}/well_{well}_electrode_{electrode}_burst_spikes.npy')
            well_spikes.append(spikedata[:,0])
            spikes_for_kde = np.append(spikes_for_kde, spikedata[:, 0])
            if len(burstdata)>0:
                burst_spikes.append(burstdata[:,0])
                burst_spikes_for_kde = np.append(burst_spikes_for_kde, burstdata[:,0])
            else:
                burst_spikes.append([])
        lineoffsets1=np.arange(1,electrode_amnt+1)
        ax[0].eventplot(well_spikes, alpha=0.5, lineoffsets=lineoffsets1)
        ax[0].eventplot(burst_spikes, alpha=0.25, color='red', lineoffsets=lineoffsets1)

        # Plot the burst frequency
        time_seconds=np.arange(0, measurements/hertz, 1/hertz)
        KDE_grid=np.arange(0, measurements/hertz, (1/hertz)*10)
        burst_freq=np.zeros(measurements)

        burst_cores_list=[]     # List that contains all the burst of an electrode

        # Load in burst core data
        for electrode in range(1,electrode_amnt+1):
            # Load in spikedata
            burst_cores=np.load(f'{burstpath}/well_{well}_electrode_{electrode}_burst_cores.npy')
            burst_cores_list.append(burst_cores)
            for burstcore in burst_cores:
                burst_freq[int(burstcore[2]):int(burstcore[3])]+=1
                    
        ax[1].plot(time_seconds, burst_freq)
        ax[0].set_title(f"Well {well}")

        data_time=measurements/hertz
        # Plot the spike frequency using kernel density estimate
        if len(spikes_for_kde)>0:
            y = FFTKDE(bw=0.1, kernel='gaussian').fit(spikes_for_kde).evaluate(grid_points=KDE_grid)
            x=KDE_grid
            y = (y - np.min(y)) / (np.max(y) - np.min(y))

            ax[2].clear()
            ax[2].plot(x,y)

        
        if len(burst_spikes_for_kde)>0:     # Check if there are bursts
            # Plot the spike frequency of spikes contained in bursts
            y = FFTKDE(bw=0.1, kernel='gaussian').fit(burst_spikes_for_kde).evaluate(grid_points=KDE_grid)
            x=KDE_grid
            y = (y - np.min(y)) / (np.max(y) - np.min(y))

            ax[3].clear()
            # Determine the threshold
            if threshold_method=='yen' or threshold_method=='Yen':
                threshold=filters.threshold_yen(y, nbins=1000)
            elif threshold_method=='otsu' or threshold_method=='Otsu':
                threshold=filters.threshold_otsu(y, nbins=1000)
            else:
                raise ValueError(f"\"{threshold_method}\" is not a valid thresholding method")
            #print(f"Threshold using {threshold_method} method set at: {threshold}")
            ax[3].plot(x,y)
            ax[3].axhline(y=threshold, color='red', linestyle='-', linewidth=1)
            # Identify regions that cross the threshold
            network_burst_cores=[]
            above_threshold=y>threshold
            i=0
            # Loop through the data
            while i < len(above_threshold):
                if above_threshold[i]:
                    # A threshold crossing has been found
                    start=x[i]
                    while above_threshold[i] and i+1<len(above_threshold):
                        i+=1
                    # When this while loop stops, the line has dropped below the threshold
                    end=x[i]
                    network_burst_cores.append([start, end])
                i+=1

            # Validate the network bursts by checking how many channels are participating
            # At the same time, remove burst cores that are too close to the beginning or end of the measurement (buffer)
            # otherwise, these will disrupt the feature calculation
            buffer=1 #second
            unvalid_burst=[]
            for i in range(len(network_burst_cores)):
                channels_participating=0
                for channel in burst_cores_list:
                    channel_participates=False
                    for burst in channel:
                        if overlap(network_burst_cores[i], burst):
                            channel_participates=True
                    if channel_participates: channels_participating+=1
                too_close = network_burst_cores[i][0]<buffer or network_burst_cores[i][1]>(measurements/hertz)-buffer
                if channels_participating<min_channels or too_close:
                    unvalid_burst.append(i)
            # Now remove all the network bursts that are not valid
            for j in sorted(unvalid_burst, reverse=True):
                del network_burst_cores[j]

            # Calculate which SCBs were participating in the network bursts, and add their extremes as the outer edges of the network burst   
            participating_bursts=[]
            for i in range(len(network_burst_cores)):
                outer_start=network_burst_cores[i][0]
                outer_end=network_burst_cores[i][1]
                for j in range(len(burst_cores_list)):
                    for burst in burst_cores_list[j]:
                        if overlap(network_burst_cores[i], burst):
                            if burst[0]<outer_start: outer_start=burst[0]
                            if burst[1]>outer_end: outer_end=burst[1]
                            # Add the identifiers (Network_burst ID, electrode, burst ID) to the list
                            participating_bursts.append([i, j, int(burst[4])])
                network_burst_cores[i].append(outer_start)
                network_burst_cores[i].append(outer_end)
                
                # Give each network burst a number (identifier)
                network_burst_cores[i].append(i)
            
            # Remove overlapping bursts
            overlapping_bursts=[]
            # Loop through all the bursts
            for i in range(len(network_burst_cores)-1):
                # Check if there is overlap with the next burst
                if overlap(network_burst_cores[i][2:4], network_burst_cores[i+1][2:4]):
                    # Mark the shortest burst for removal
                    if (network_burst_cores[i][1]-network_burst_cores[i][0])<(network_burst_cores[i+1][1]-network_burst_cores[i+1][0]):
                        overlapping_bursts.append(i)
                    else:
                        overlapping_bursts.append(i+1)
            if len(overlapping_bursts)>0:
                # Remove the overlapping bursts (we do this outside of the loop to not disturb the loop)
                for j in sorted(overlapping_bursts, reverse=True):
                    del network_burst_cores[j]
                # Remove participating bursts with the removed ID
                participating_bursts=np.array(participating_bursts)[~np.isin(np.array(participating_bursts)[:,0],overlapping_bursts)]

            total_network_burst=(network_burst_cores)
            # Highlight the burst cores in every graph except te top one
            for graph in ax[:]:
                for network_burst in network_burst_cores:
                    temp=np.array(network_burst[:4])
                    graph.axvspan(temp[0], temp[1], facecolor='green', alpha=0.75)
                    graph.axvspan(temp[2], temp[3], facecolor='green', alpha=0.5)
                  
            ax[2].set_ylabel("Spike\nactivity\n(KDE)")
            ax[3].set_ylabel("Burst\nspike\nactivity\n(KDE)")
        else:
            total_network_burst=[]
            participating_bursts=[]
        if savedata:
            path = f'{outputpath}/network_data'
            np.savetxt(f'{path}/well_{well}_network_bursts.csv', total_network_burst, delimiter = ",")
            np.save(f'{path}/well_{well}_network_bursts', total_network_burst)
            np.savetxt(f'{path}/well_{well}_participating_bursts.csv', participating_bursts, delimiter = ",")
            np.save(f'{path}/well_{well}_participating_bursts', participating_bursts)
    if plot_electrodes:
        # Save the figure
        path=f"{outputpath}/figures/well_{well}"
        fig.savefig(path, dpi=120)
        return fig
    else:
        plt.cla()
        plt.clf()


import numpy as np

'''Threshold function - returns threshold value'''
def threshold(data, hertz, stdevmultiplier, RMSmultiplier, threshold_portion):
    measurements=data.shape[0]
    # The amount is samples needed to for a 50ms window is calculated
    windowsize = 0.05 * hertz
    windowsize = int(windowsize)
    # Create a temporary list that will contain 50ms windows of data
    windows = []

    # Iterate over the electrode data and create x amount of windows containing *windowsize* samples each
    # For this it is pretty important that the data consists of a multitude of 50ms seconds
    for j in range(0, int(measurements), windowsize):
        windows.append(data[j:j+windowsize])
    windows = np.array(windows) #convert into np.array
    # Create an empty list where all the data identified as spike-free noise will be stored
    noise=[]

    # Now iterate over every 50ms time window and check whether it is "spike-free noise"
    # If all of the electrode should be used to determine the noise window
    if threshold_portion==1:
        loopvalues=range(0,windows.shape[0])
    # If this is not the case, only use a portion of the data to determine the noise level
    else:
        loopvalues=range(0,windows.shape[0])
        step=int(1/threshold_portion)
        loopvalues=loopvalues[::step]
    for j in loopvalues:
        # Calculate the mean and standard deviation
        #mu, std = norm.fit(windows[j])
        std=np.std(windows[j])
        # Check whether the minimum/maximal value lies outside the range of x (defined above) times the standard deviation - if this is not the case, this 50ms box can be seen as 'spike-free noise'
        if not(np.min(windows[j])<(-1*stdevmultiplier*std) or np.max(windows[j])>(stdevmultiplier*std)):
            # 50ms has been identified as noise and will be added to the noise paremeter
            noise = np.append(noise, windows[j][:])

    # Calculate the RMS
    RMS=np.sqrt(np.mean(noise**2))
    threshold_value=RMSmultiplier*RMS

    # Calculate the % of the file that was noise
    noisepercentage=noise.shape[0]/data.shape[0]
    #print(f'{noisepercentage*100}% of data identified as noise')
    return threshold_value

def fast_threshold(data, hertz, stdevmultiplier, RMSmultiplier, threshold_portion):
    measurements=data.shape[0]
    # The stepsize needed for a 50ms window is calculated
    stepsize = int(0.05 * hertz)
    total_steps=int(measurements/stepsize)

    loopvalues=range(0,total_steps)
    step=int(1/threshold_portion)
    loopvalues=loopvalues[::step]
    noise=np.array([])

    for j in loopvalues:
        # Calculate the mean and standard deviation
        #mu, std = norm.fit(windows[j])
        std=np.std(data[j*stepsize:(j+1)*stepsize])
        # Check whether the minimum/maximal value lies outside the range of x (defined above) times the standard deviation - if this is not the case, this 50ms box can be seen as 'spike-free noise'
        if not(np.min(data[j*stepsize:(j+1)*stepsize])<(-1*stdevmultiplier*std) or np.max(data[j*stepsize:(j+1)*stepsize])>(stdevmultiplier*std)):
            # 50ms has been identified as noise and will be added to the noise paremeter
            noise = np.append(noise, data[j*stepsize:(j+1)*stepsize])

    # Calculate the RMS
    RMS=np.sqrt(np.mean(noise**2))
    threshold_value=RMSmultiplier*RMS

    # Calculate the % of the file that was noise
    noisepercentage=noise.shape[0]/measurements
    #print(f'{noisepercentage*100}% of data identified as noise')
    return threshold_value


import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from functools import partial
from KDEpy import FFTKDE
import seaborn

'''Create raster plots for all the electrodes'''
def raster(electrodes, electrode_amnt, samples, hertz, filename):
    # Check which electrodes are given, and how these will be plotted
    i=0
    
    filename=filename[:-3]
    filename=filename+"_values"
    while i < len(electrodes):
        well_spikes=[]
        burst_spikes=[]
        # Collect all the data from a single well
        while i<len(electrodes): # and ((electrodes[i])%electrode_amnt)!=0:
            electrode = electrodes[i] % electrode_amnt + 1
            well = round(electrodes[i] / electrode_amnt + 0.505)
            path=f'spike_values'
            spikedata=np.load(f'{filename}/{path}/well_{well}_electrode_{electrode}_spikes.npy')
            path='burst_values'
            burstdata=np.load(f'{filename}/{path}/well_{well}_electrode_{electrode}_burst_spikes.npy')
            well_spikes.append(spikedata[:,0])
            if len(burstdata)>0:
                burst_spikes.append(burstdata[:,0])
            else:
                burst_spikes.append([])
            if ((electrodes[i]+1)%electrode_amnt)==0: break
            i+=1
        amount_of_electrodes=len(burst_spikes)
        plt.cla()
        plt.rcParams["figure.figsize"] = (22,5)
        end_electrode=((electrodes[i-1]+1)%12)+1
        start_electrode = end_electrode-amount_of_electrodes
        lineoffsets1=np.arange(start_electrode+1, end_electrode+1)
        plt.eventplot(well_spikes, alpha=0.5, lineoffsets=lineoffsets1)
        plt.eventplot(burst_spikes, alpha=0.5, color='red', lineoffsets=lineoffsets1)
        plt.xlim([0,samples/hertz])
        plt.ylim([start_electrode, end_electrode+1])
        plt.yticks(lineoffsets1)
        plt.title(f"Well {well}")
        plt.xlabel("Time in seconds")
        plt.ylabel("Electrode")
        plt.show()
        i+=1

'''This function will calculate a 3d guassian kernel'''
def K(x, H):
    # unpack two dimensions
    x1, x2 = x
    # extract four components from the matrix inv(H)
    a, b, c, d = np.linalg.inv(H).flatten()
    
    scale = 2*np.pi*np.sqrt( np.linalg.det(H))
    return np.exp(-(a*x1**2 + d*x2**2 + (b+c)*x1*x2)/2) / scale

'''This function will insert a 3d gaussian kernel at the location of every datapoint/spike'''
def KDE(x, H, data):
    # unpack two dimensions
    x1, x2 = x
    # prepare the grid for output values
    output = np.zeros_like(x1)
    # process every sample
    for sample in data:
        output += K([x1-sample[0], x2-sample[1]], H)
    return output

'''Create a 3D view of a single well. Every spike will be represented as a 3D gaussian kernel'''
def fancyplot(filename, wells, electrode_amnt, measurements, hertz, resolution, kernel_size, aspectratios, colormap):
    # Convert the filename so the algorithm can look through the folders
    filename=filename[:-3]
    filename=filename+"_values"
    spikepath=f'{filename}/spike_values'
    for well in wells:
        kdedata=[]
        for electrode in range(1,electrode_amnt+1):
            # Load in spikedata
            spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode}_spikes.npy')
            spikedata=spikedata[:,:2]
            spikedata[:,1]=electrode
            for spike in spikedata:
                kdedata.append(spike)
        time=measurements/hertz

        # covariance matrix
        # This determines the shape of the gaussian kernel
        H = [[1, 0],
            [0, 1]]

        data = np.array(kdedata)
        # fix arguments 'H' and 'data' of KDE function for further calls
        KDE_partial = partial(KDE, H=kernel_size*np.array(H), data=data)

        # draw contour and surface plots
        func=KDE_partial

        # create a np xy grid using the dimensions of the data
        yres=int(resolution*time)
        xres=int(resolution*electrode_amnt)
        y_range = np.linspace(start=0, stop=time, num=yres)
        x_range = np.linspace(start=0, stop=electrode_amnt, num=xres)
        print(f"Creating 3D plot, resolution x: {xres} by y: {yres}")

        y_grid, X_grid = np.meshgrid(y_range, x_range)
        Z_grid = func([y_grid, X_grid])
        
        fig = go.Figure(data=[go.Surface(y=y_grid, x=X_grid, z=Z_grid, colorscale=colormap)])
        fig.update_layout(
        scene = dict(
            xaxis = dict(range=[1,electrode_amnt]),
            yaxis = dict(range=[0,time]),
            ),
        margin=dict(r=20, l=10, b=10, t=10))

        # change the aspect ratio's of the plot
        fig.update_layout(scene_aspectmode='manual',
                            title=f"Well: {well}",
                            scene_aspectratio=dict(x=(electrode_amnt/10)*aspectratios[0],
                                           y=(time/10)*aspectratios[1],
                                           z=1*aspectratios[2]))
        fig.update_layout(scene = dict(
                    xaxis_title='Electrode',
                    yaxis_title='Time in seconds',
                    zaxis_title='KDE'),)
        return fig
        # fig.show()
        # fig.write_html(f"{filename}/well_{well}_graph.html")

'''Plots the KDE of all electrodes (x) of a well in x amount of subplots'''
def well_electrodes_kde(outputpath, well, electrode_amnt, measurements, hertz, bandwidth=1):
    spikepath=f'{outputpath}/spike_values'
    data_time=measurements/hertz
    fig, ax = plt.subplots(electrode_amnt, 1, sharex=True, sharey=True)
    for electrode in range(electrode_amnt):
        spikedata=np.load(f'{spikepath}/well_{well}_electrode_{electrode+1}_spikes.npy')
        if len(spikedata[:,0])>0:
            y = FFTKDE(bw=bandwidth, kernel='gaussian').fit(spikedata[:,0]).evaluate(grid_points=np.arange(0, data_time, 0.001))
            y=np.array(y)*len(spikedata[:,0])
            x=np.arange(0, data_time, 0.001)
        else:
            x=np.arange(0, data_time, 0.001)
            y=np.zeros(len(x))
        ax[electrode].plot(x,y)
        ax[electrode].set_ylabel(f"E: {electrode+1}")
    ax[-1].set_xlabel("Time in seconds")
    ax[0].set_xlim([0, measurements/hertz])
    ax[0].set_title(f"Well: {well} activity")
    return fig


from CureQ.bandpass import *
from CureQ.burst_detection import *
from CureQ.features import *
from CureQ.network_burst_detection import *
from CureQ.open_file import *
from CureQ.plotting import *
from CureQ.spike_validation import *
from CureQ.threshold import *
import time
import multiprocessing
from multiprocessing.managers import SharedMemoryManager
import gc
import json
from datetime import date
from datetime import datetime

'''Analyse electrode as subprocess'''
def _electrode_subprocess(outputpath, memory_id, shape, _type, electrode, hertz, low_cutoff, high_cutoff, order, stdevmultiplier,
                        RMSmultiplier, threshold_portion, spikeduration, exit_time_s,
                        amplitude_drop_sd, plot_electrodes, electrode_amnt,
                        heightexception, max_boxheight, kde_bandwidth, smallerneighbours,
                        minspikes_burst, maxISI_outliers, default_threshold,
                        validation_method, progressfile):
    # Attach to the existing shared memory block
    existing_shm = multiprocessing.shared_memory.SharedMemory(name=memory_id)
    funcdata=np.ndarray(shape, _type, buffer=existing_shm.buf)
    data=funcdata[electrode]

    # Filter the data
    data=butter_bandpass_filter(data, low_cutoff, high_cutoff, hertz, order)
    
    # Calculate the threshold
    threshold_value=fast_threshold(data, hertz, stdevmultiplier, RMSmultiplier, threshold_portion)
    
    # Calculate spike values
    if validation_method=="DMP_noisebased":
        spike_validation(data, electrode, threshold_value, hertz, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, heightexception, max_boxheight, outputpath)
    elif validation_method=='none':
        spike_validation(data, electrode, threshold_value, hertz, spikeduration, exit_time_s, 0, plot_electrodes, electrode_amnt, heightexception, max_boxheight, outputpath)
    else:
        raise ValueError(f"\"{validation_method}\" is not a valid spike validation method")
    
    # Detect the bursts
    burst_detection(data, electrode, electrode_amnt, hertz, kde_bandwidth, smallerneighbours, minspikes_burst, maxISI_outliers, default_threshold, outputpath, plot_electrodes)
    print(f"Calculated electrode: {electrode}")

'''Analyse an entire well'''
def analyse_well(fileadress,                                # Where is the data file stored
                      hertz,                                # What is the sampling frequency of the MEA
                      electrode_amnt,                       # The amount of electrodes per well
                      wells='all',                          # Which wells do you want to analyze
                      validation_method="DMP_noisebased",   # Which validation method do you want to use, possible: "DMP", "DMP_noisebased"
                      low_cutoff=200,                       # The low_cutoff for the bandpass filter
                      high_cutoff=3500,                     # The high_cutoff for the bandpass filter
                      order=2,                              # The order for the bandpass filter
                      spikeduration=0.002,                  # The amount of time only 1 spike should be registered, aka refractory period
                      exit_time_s=0.00024,                  # The amount of time a spike gets to drop amplitude in the validation methods
                      plot_electrodes=True,                 # Plot the single electrode visualizations
                      well_amnt=24,                         # The amount of wells present in the MEA
                      kde_bandwidth=1,                      # The bandwidth of the kernel density estimate
                      smallerneighbours=10,                 # The amount of smaller neighbours a peak should have before being considered as one
                      minspikes_burst=5,                    # The minimal amount of spikes a burst should have
                      maxISI_outliers=1000,                 # The maximal ISIth2 that can be used in burst detection
                      default_threshold=100,                # The default value for ISIth1
                      heightexception=1.5,                  # Multiplied with the spike detection threshold, if a spike exceeds this value, it does not have to drop amplitude fast to be validated
                      max_boxheight=2,                      # Multiplied with the spike detection threshold, will be the maximal value the box for DMP_noisebased validation can be
                      amplitude_drop_sd=5,                  # Multiplied with the SD of surrounding noise, will be the boxheight for DMP_noisebased validation
                      stdevmultiplier=5,                    # The amount of SD a value needs to be from the mean to be considered a possible spike in the threshold detection
                      RMSmultiplier=5,                      # Multiplied with the RMS of the spike-free noise, used to determine the threshold
                      min_channels=0.5,                     # Minimal % of channels that should participate in a burst
                      threshold_method='yen',               # Threshold method to decide whether activity is a network burst or not - possible: 'yen', 'otsu'
                      
                      # Parameters for the 3D plot
                    #   resolution=5,                         # Creates a resolution*electrode_amnt by resolution*time_seconds grid for the 3D plot. E.g. a measurement time of 150s with 12 electrodes and resolution 10 would give a 1500*120 grid. Higher resolution means longer computing times
                    #   kernel_size=1,                        # The size of the 3D gaussian kernel
                    #   aspectratios=[0.5,0.25,0.5],          # The aspect ratios of the plot. multiplied with the length of the xyz axis'.
                    #   colormap="deep",                      # Colormap of the 3D plot

                      spikes_are_analysed=False,            # If this values is true, the algorithm will search for the spikefiles instead of recalculating these values
                      activity_threshold=0.1,               # The lowest frequency an electrode can have before being removed from the analysis
                      threshold_portion=0.1,                  # How much of the electrode do you want to use to calculate the threshold. Higher number = higher runtime
                      remove_inactive_electrodes=True,      # Whether you want to remove inactive electrodes

                      # Parameters for cutting up the data
                      cut_data_bool=False,
                      parts=10,
                      recordingtime=120,

                      use_multiprocessing=False
                      ):

    # Advertisement
    print("MEA-analyzer. Developed by Joram van Beem and Jesse Antonissen for the CureQ research consortium")
    
    # Create a directory which will contain the output
    outputpath=os.path.splitext(fileadress)[0]#remove the file extension
    outputpath=outputpath+"_output"
    # Add current datetime
    outputpath=outputpath+f"_{date.today()}"
    now = datetime.now()
    current_time = now.strftime("%H-%M-%S")
    outputpath=outputpath+f"_{current_time}"
    os.makedirs(outputpath)

    # Create different folders for the output
    os.makedirs(f"{outputpath}/burst_values")
    os.makedirs(f"{outputpath}/spike_values")
    os.makedirs(f"{outputpath}/network_data")
    os.makedirs(f"{outputpath}/figures")


    # Create a file to commmunicate the progress with the GUI
    progressfile=f'{os.path.split(fileadress)[0]}/progress.npy'
    np.save(progressfile, ['starting'])

    if not spikes_are_analysed:
        print("Opening the data")
        data=openHDF5_SCA1(fileadress)
    if wells=='all':
        wells=list(range(1,int(data.shape[0]/electrode_amnt)+1))
    if cut_data_bool:
        print("Dividing the data")
        data=cut_data(data, parts, electrode_amnt)
        new_wells=np.array([])
        for well in wells:
            temp=(np.arange((well-1)*parts+1, (well)*(parts)+1))
            new_wells=np.append(new_wells, temp)
        wells=new_wells.astype(int)
    first_iteration=True

    # Save the parameters that have been given in a JSON file 
    parameters={
        'file adress' : fileadress,
        'wells' : wells,
        'sampling rate' : hertz,
        'electrode amount' : electrode_amnt,
        'low cutoff' : low_cutoff,
        'high cutoff' : high_cutoff,
        'order' : order,
        'threshold portion' : threshold_portion,
        'standard deviation multiplier' : stdevmultiplier,
        'rms multiplier' : RMSmultiplier,
        'refractory period' : spikeduration,
        'spike validation method' : validation_method,
        'exit time' : exit_time_s,
        'drop amplitude' : amplitude_drop_sd,
        'height exception' : heightexception,
        'max drop amount' : max_boxheight,
        'minimal amount of spikes' : minspikes_burst,
        'default interval threshold' : default_threshold,
        'max interval threshold' : maxISI_outliers,
        'KDE bandwidth' : kde_bandwidth,
        'smaller neighbours' : smallerneighbours,
        'min channels' : min_channels,
        'thresholding method' : threshold_method,
        'remove inactive electrodes' : remove_inactive_electrodes,
        'activity threshold' : activity_threshold,
        'split data' : cut_data_bool,
        'parts' : parts,
        'use multiprocessing' : use_multiprocessing,
        'measurements' : data.shape[1]
    }
    with open(f"{outputpath}/parameters.json", 'w') as outfile:
        json.dump(parameters, outfile)

    # With multiprocessing
    if use_multiprocessing:
        # Save the data in shared memory
        sharedmemory=multiprocessing.shared_memory.SharedMemory(create=True, size=data.nbytes)
        shape=data.shape
        np.save(progressfile, [(0)*electrode_amnt, shape[0]])
        _type=data.dtype
        data_shared=np.ndarray(shape, dtype=_type, buffer=sharedmemory.buf)
        data_shared[:]=data[:]
        memory_id=sharedmemory.name
        measurements=data.shape[1]
        data=[] # hopefully this clears up memory space

        # Start up a process for every single electrode
        with multiprocessing.Pool(processes=electrode_amnt) as pool:
            for well in wells:
                start=time.time()
                electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
                print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

                # Divide the tasks to the processes
                args=[(outputpath, memory_id, shape, _type, electrode, hertz, low_cutoff, high_cutoff, order, stdevmultiplier, RMSmultiplier, threshold_portion, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, heightexception, max_boxheight, kde_bandwidth, smallerneighbours, minspikes_burst, maxISI_outliers, default_threshold, validation_method, progressfile) for electrode in electrodes]
                pool.starmap(_electrode_subprocess, args)

                # Calculate the network bursts
                network_burst_detection(outputpath, [well], electrode_amnt, measurements, hertz, min_channels, threshold_method, plot_electrodes)

                features_df=electrode_features(outputpath, electrodes, electrode_amnt, measurements, hertz, activity_threshold, remove_inactive_electrodes)
                well_features_df=well_features(outputpath, well, electrode_amnt, measurements, hertz)
                if first_iteration:
                    first_iteration=False
                    output=feature_output(features_df, well_features_df, electrode_amnt)
                else:
                    output=pd.concat([output, feature_output(features_df, well_features_df, electrode_amnt)], axis=0)
                end=time.time()
                print(f"It took {end-start} seconds to analyse well: {well}")
                np.save(progressfile, [(well)*electrode_amnt, shape[0]])
        # Clean up the shared memory
        sharedmemory.close()
        sharedmemory.unlink()
    # Without multiprocessing
    else:
        for well in wells:
            start=time.time()
            electrodes=np.arange((well-1)*electrode_amnt, well*electrode_amnt)
            print(f"Analyzing well: {well}, consisting of electrodes: {electrodes}")

            # We don't have to analyse the spikes again if the files are already there
            if not spikes_are_analysed:
                # Loop through all the electrodes
                for electrode in electrodes:
                    # Filter the data
                    data[electrode]=butter_bandpass_filter(data[electrode], low_cutoff, high_cutoff, hertz, order)
                    
                    # Calculate the threshold
                    threshold_value=fast_threshold(data[electrode], hertz, stdevmultiplier, RMSmultiplier, threshold_portion)
                    
                    # Calculate spike values
                    if validation_method=="DMP_noisebased":
                        spike_validation(data[electrode], electrode, threshold_value, hertz, spikeduration, exit_time_s, amplitude_drop_sd, plot_electrodes, electrode_amnt, heightexception, max_boxheight, outputpath)
                    elif validation_method=='none':
                        spike_validation(data[electrode], electrode, threshold_value, hertz, spikeduration, exit_time_s, 0, plot_electrodes, electrode_amnt, heightexception, max_boxheight, outputpath)
                    else:
                        raise ValueError(f"\"{validation_method}\" is not a valid spike validation method")
                    
                    # Detect the bursts
                    burst_detection(data[electrode], electrode, electrode_amnt, hertz, kde_bandwidth, smallerneighbours, minspikes_burst, maxISI_outliers, default_threshold, outputpath, plot_electrodes)
                    np.save(progressfile, [electrode+1, data.shape[0]])
                    print(f"Calculated electrode: {electrode}")
            if spikes_are_analysed:
                measurements=recordingtime*hertz
            else:
                measurements=data.shape[1]

            network_burst_detection(outputpath, [well], electrode_amnt, measurements, hertz, min_channels, threshold_method, plot_electrodes)
            print(f"Calculated network bursts well: {well}")

            features_df=electrode_features(outputpath, electrodes, electrode_amnt, measurements, hertz, activity_threshold, remove_inactive_electrodes)
            well_features_df=well_features(outputpath, well, electrode_amnt, measurements, hertz)
            if first_iteration:
                first_iteration=False
                output=feature_output(features_df, well_features_df, electrode_amnt)
            else:
                output=pd.concat([output, feature_output(features_df, well_features_df, electrode_amnt)], axis=0, ignore_index=False)
            end=time.time()
            print(f"It took {end-start} seconds to analyse well: {well}")
        
        # Attempt to free up RAM
        data=[]
        del data
        gc.collect()
    name = os.path.basename(outputpath)
    name = os.path.splitext(name)
    output.to_csv(f"{outputpath}/Features.csv", index=False)
    np.save(progressfile, ['done'])
    return output